# LEGACY MODULE - TODO: remove or refactor

import abc

import numpy as np
from scipy.integrate import simps
from sklearn.base import BaseEstimator
from sklearn.neighbors import KernelDensity

from skpro.utils.utils import not_existing


def ecdf(a):
    """Returns the empirical distribution function of a sample

    Parameters
    ----------
    a: array
        Input array representing a sample

    Returns
    -------
    array xs   Empirical cdf of the input sample
    array ys
    """
    xs = np.sort(np.array(a))
    ys = np.arange(1, len(xs) + 1) / float(len(xs))

    return xs, ys


def step_function(xs, ys):
    """
    Returns a step function from x-y pair sample

    Parameters
    ----------
    xs  x values
    ys  corresponding y values

    Returns
    -------
    function A step function
    """

    def func(x):
        index = np.searchsorted(xs, x)
        index = len(ys) - 1 if index >= len(ys) else index
        return ys[index]

    return func


class DensityAdapter(BaseEstimator, metaclass=abc.ABCMeta):
    """
    Abstract base class for density adapter
    that transform an input into an
    density cdf/pdf interface
    """

    @abc.abstractmethod
    def __call__(self, inlet):
        """
        Adapter entry point

        Parameters
        ----------
        mixed inlet Input for the adapter transformation
        """
        raise NotImplementedError()

    @abc.abstractmethod
    def pdf(self, x):
        """Probability density function

        Parameters
        ----------
        x

        Returns
        -------
        mixed  Density function evaluated at x
        """
        raise NotImplementedError()

    @abc.abstractmethod
    def cdf(self, x):
        """Cumulative density function

        Parameters
        ----------
        x

        Returns
        -------
        mixed Cumulative density function evaluated at x
        """
        raise NotImplementedError()


class KernelDensityAdapter(DensityAdapter):
    """
    DensityAdapter that uses kernel density estimation
    to transform samples
    """

    def __init__(self, estimator=KernelDensity()):
        self.estimator = estimator

    def __call__(self, sample):
        """
        Adapter entry point

        Parameters
        ----------
        np.array(M) inlet: Sample of length M
        """

        # fit kernel density estimator
        self._min_sample = min(sample)
        self._max_sample = max(sample)
        self._std_sample = np.std(sample)

        self.estimator.fit(sample[:, np.newaxis])

    def cdf(self, x):
        a = 10
        grid_size = 1000

        minus_inf = self._min_sample - a * self._std_sample

        step = (x - minus_inf) / grid_size
        grid = np.arange(minus_inf, x, step)

        pdf_estimation = np.exp(self.estimator.score_samples(grid.reshape(-1, 1)))
        integral = simps(y=pdf_estimation, dx=step)

        return integral

    def pdf(self, x):
        x = np.array(x)
        try:
            return np.exp(self.estimator.score_samples(x))[0]
        except ValueError:
            return np.exp(self.estimator.score_samples(x.reshape(-1, 1)))[0]


class EmpiricalDensityAdapter(DensityAdapter):
    """
    DensityAdapter that uses empirical cdf
    to transform samples
    """

    def __init__(self):
        self.xs_ = None
        self.ys_ = None
        self.step_function_ = None

    def __call__(self, sample):
        """
        Adapter entry point

        Parameters
        ----------
        np.array(M) inlet: Bayesian sample of length M
        """
        self.xs_, self.ys_ = ecdf(sample)
        self.step_function_ = step_function(self.xs_, self.ys_)

    def cdf(self, x):
        return self.step_function_(x)

    @not_existing
    def pdf(self, x):
        pass



# LEGACY MODULE - TODO: remove or refactor

if False:
    import numpy as np
    from hypothesis import given
    from hypothesis.extra.numpy import arrays
    from hypothesis.strategies import floats
    from scipy.stats import norm

    from skpro.density import EmpiricalDensityAdapter, KernelDensityAdapter, ecdf

    np.random.seed(1)

    @given(arrays(np.float, 10, elements=floats(0, 100)))
    def test_ecdf_from_sample(sample):
        xs, ys = ecdf(sample)

        # correct mapping?
        assert len(xs) == len(ys)

        # is it monotone?
        assert np.array_equal(ys, sorted(ys))

    @given(floats(-10, 10))
    def test_kernel_density_adapter(x):
        # Bayesian test sample
        loc, scale = 5, 10
        sample = np.random.normal(loc=loc, scale=scale, size=500)

        # Initialise adapter
        adapter = KernelDensityAdapter()
        adapter(sample)

        # PDF
        pdf = adapter.pdf(x)
        assert isinstance(pdf, np.float)
        assert abs(pdf - norm.pdf(x, loc=loc, scale=scale)) < 0.3

        # CDF
        cdf = adapter.cdf(x)
        assert isinstance(cdf, np.float)
        assert abs(cdf - norm.cdf(x, loc=5, scale=10)) < 0.3

    @given(floats(-10, 10))
    def test_empirical_density_adapter(x):
        # Bayesian test sample
        loc, scale = 5, 10

        sample = np.random.normal(loc=loc, scale=scale, size=5000)

        # Initialise adapter
        adapter = EmpiricalDensityAdapter()
        adapter(sample)

        # CDF
        cdf = adapter.cdf(x)
        assert isinstance(cdf, float)
        assert abs(cdf - norm.cdf(x, loc=loc, scale=scale)) < 0.3

        # PDF
        assert adapter.pdf.not_existing



"""LEGACY MODULE - TODO: remove or refactor."""

import pytest
from sklearn.ensemble import BaggingRegressor as ClassicBaggingRegressor
from sklearn.metrics import mean_squared_error as mse
from sklearn.tree import DecisionTreeRegressor

from skpro.workflow.manager import DataManager


def prediction(model, data):
    return model.fit(data.X_train, data.y_train).predict(data.X_test)


@pytest.mark.skip(reason="loss assert fails sporadically")
def test_bagging_wrapper():
    data = DataManager("boston")

    # Run classic sklearn bagging mechanism to ensure we have
    # the correct parameters in place
    baseline_classic = prediction(DecisionTreeRegressor(), data)

    bagged_classic = prediction(ClassicBaggingRegressor(DecisionTreeRegressor()), data)
    #
    # # Does the bagging reduce the loss?
    assert mse(data.y_test, baseline_classic) > mse(data.y_test, bagged_classic)

    # Run corresponding skpro bagging mechanism

    # clf = DecisionTreeRegressor()
    #
    # baseline_prediction = prediction(
    #     ParametricEstimator(point=clf),
    #     data
    # )
    #
    # skpro_bagging_prediction = prediction(
    #     SkproBaggingRegressor(
    #         ParametricEstimator(point=clf),
    #         n_estimators=10,
    #         n_jobs=-1
    #     ),
    #     data
    # )
    #
    # l1, l2 = loss(data.y_test, baseline_prediction),\
    #          loss(data.y_test, skpro_bagging_prediction)
    #
    # # Does the bagging reduce the loss?
    # assert l1 > l2



# copyright: sktime developers, BSD-3-Clause License (see LICENSE file)
"""Common utilities for polars based data containers."""
from skpro.datatypes._common import _req
from skpro.datatypes._common import _ret as ret


def check_polars_frame(obj, return_metadata=False, var_name="obj", lazy=False):
    """Check polars frame, generic format."""
    import polars as pl

    metadata = {}

    if lazy:
        exp_type = pl.LazyFrame
        exp_type_str = "LazyFrame"
    else:
        exp_type = pl.DataFrame
        exp_type_str = "DataFrame"

    if not isinstance(obj, exp_type):
        msg = f"{var_name} must be a polars {exp_type_str}, found {type(obj)}"
        return ret(False, msg, None, return_metadata)

    # we now know obj is a polars DataFrame or LazyFrame
    if _req("is_empty", return_metadata):
        metadata["is_empty"] = obj.width < 1
    if _req("is_univariate", return_metadata):
        metadata["is_univariate"] = obj.width == 1
    if _req("n_instances", return_metadata):
        if hasattr(obj, "height"):
            metadata["n_instances"] = obj.height
        else:
            metadata["n_instances"] = "NA"
    if _req("n_features", return_metadata):
        metadata["n_features"] = obj.width
    if _req("feature_names", return_metadata):
        metadata["feature_names"] = obj.columns

    # check if there are any nans
    #   compute only if needed
    if _req("has_nans", return_metadata):
        if isinstance(obj, pl.LazyFrame):
            metadata["has_nans"] = "NA"
        else:
            hasnan = obj.null_count().sum_horizontal().to_numpy()[0] > 0
            metadata["has_nans"] = hasnan

    return ret(True, None, metadata, return_metadata)
